@@ -8,11 +8,11 @@
 */
 
 #include "knotification.h"
+#include "knotification_p.h"
 #include "knotificationmanager_p.h"
 
 #include <config-knotifications.h>
 
-#include <KPluginLoader>
 #include <KPluginMetaData>
 #include <QFileInfo>
 #include <QHash>
@@ -115,6 +115,7 @@ KNotificationPlugin *KNotificationManager::pluginForAction(const QString &action
     auto addPlugin = [this](KNotificationPlugin *plugin) {
         d->notifyPlugins[plugin->optionName()] = plugin;
         connect(plugin, &KNotificationPlugin::finished, this, &KNotificationManager::notifyPluginFinished);
+        connect(plugin, &KNotificationPlugin::xdgActivationTokenReceived, this, &KNotificationManager::xdgActivationTokenReceived);
         connect(plugin, &KNotificationPlugin::actionInvoked, this, &KNotificationManager::notificationActivated);
         connect(plugin, &KNotificationPlugin::replied, this, &KNotificationManager::notificationReplied);
     };
@@ -163,44 +164,50 @@ KNotificationPlugin *KNotificationManager::pluginForAction(const QString &action
     } else {
         bool pluginFound = false;
 
-        QList<QObject *> plugins = KPluginLoader::instantiatePlugins(
-            QStringLiteral("knotification/notifyplugins"),
-            [&action, &pluginFound](const KPluginMetaData &data) {
-                // KPluginLoader::instantiatePlugins loops over the plugins it
-                // found and calls this function to determine whether to
-                // instantiate them. We use a `pluginFound` var outside the
-                // lambda to break out of the loop once we got a match.
-                // The reason we can't just use KPluginLoader::findPlugins,
-                // loop over the meta data and instantiate only one plugin
-                // is because the X-KDE-KNotification-OptionName field is
-                // optional (see TODO note below) and the matching plugin
-                // may be among the plugins which don't have it.
-                if (pluginFound) {
+        std::function<bool(const KPluginMetaData &)> filter = [&action, &pluginFound](const KPluginMetaData &data) {
+            // KPluginMetaData::findPlugins loops over the plugins it
+            // finds and calls this function to determine whether to
+            // deliver them. We use a `pluginFound` var outside the
+            // lambda to break out of the loop once we got a match.
+            // The reason we can't just have KPluginMetaData::findPlugins,
+            // loop over the meta data and instantiate only one plugin
+            // is because the X-KDE-KNotification-OptionName field is
+            // optional (see TODO note below) and the matching plugin
+            // may be among the plugins which don't have it.
+            if (pluginFound) {
+                return false;
+            }
+
+            const QJsonObject &rawData = data.rawData();
+
+            // This field is new-ish and optional. If it's not set we always
+            // instantiate the plugin, unless we already got a match.
+            // TODO KF6: Require X-KDE-KNotification-OptionName be set and
+            // reject plugins without it.
+            if (rawData.contains(QLatin1String("X-KDE-KNotification-OptionName"))) {
+                if (rawData.value(QStringLiteral("X-KDE-KNotification-OptionName")) == action) {
+                    pluginFound = true;
+                } else {
                     return false;
                 }
+            }
 
-                const QJsonObject &rawData = data.rawData();
+            return true;
+        };
 
-                // This field is new-ish and optional. If it's not set we always
-                // instantiate the plugin, unless we already got a match.
-                // TODO KF6: Require X-KDE-KNotification-OptionName be set and
-                // reject plugins without it.
-                if (rawData.contains(QLatin1String("X-KDE-KNotification-OptionName"))) {
-                    if (rawData.value(QStringLiteral("X-KDE-KNotification-OptionName")) == action) {
-                        pluginFound = true;
-                    } else {
-                        return false;
-                    }
-                }
-
-                return true;
-            },
-            this);
-
-        for (QObject *pluginObj : qAsConst(plugins)) {
+        QPluginLoader loader;
+        const QVector<KPluginMetaData> listMetaData = KPluginMetaData::findPlugins(QStringLiteral("knotification/notifyplugins"), filter);
+        for (const KPluginMetaData &metadata : listMetaData) {
+            loader.setFileName(metadata.fileName());
+            QObject *pluginObj = loader.instance();
+            if (!pluginObj) {
+                qCWarning(LOG_KNOTIFICATIONS).nospace() << "Could not instantiate plugin \"" << metadata.fileName() << "\": " << loader.errorString();
+                continue;
+            }
             KNotificationPlugin *notifyPlugin = qobject_cast<KNotificationPlugin *>(pluginObj);
 
             if (notifyPlugin) {
+                notifyPlugin->setParent(this);
                 // We try to avoid unnecessary instantiations (see above), but
                 // when they happen keep the resulting plugins around.
                 addPlugin(notifyPlugin);
@@ -242,6 +249,16 @@ void KNotificationManager::notificationActivated(int id, int action)
     }
 }
 
+void KNotificationManager::xdgActivationTokenReceived(int id, const QString &token)
+{
+    KNotification *n = d->notifications.value(id);
+    if (n) {
+        qCDebug(LOG_KNOTIFICATIONS) << "Token received for" << id << token;
+        n->d->xdgActivationToken = token;
+        Q_EMIT n->xdgActivationTokenChanged();
+    }
+}
+
 void KNotificationManager::notificationReplied(int id, const QString &text)
 {
     if (KNotification *n = d->notifications.value(id)) {
@@ -370,7 +387,7 @@ void KNotificationManager::update(KNotification *n)
 {
     KNotifyConfig notifyConfig(n->appName(), n->contexts(), n->eventId());
 
-    for (KNotificationPlugin *p : qAsConst(d->notifyPlugins)) {
+    for (KNotificationPlugin *p : std::as_const(d->notifyPlugins)) {
         p->update(n, &notifyConfig);
     }
 }
